from setuptools import setup, find_packages

setup(
    name='harmonai-tools',
    version='0.0.4',
    url='https://github.com/Harmonai-org/harmonai-tools.git',
    author='Harmonai',
    description='Training and inference tools for generative audio models from Harmonai',
    packages=find_packages(),  
    install_requires=[
        'aeiou',
        'alias-free-torch',
        'audio-diffusion-pytorch-fork',
        'audio-metadata',
        'auraloss',
        'descript-audio-codec==0.0.3',
        'einops',
        'ema-pytorch',
        'encodec',
        'gradio',
        'importlib-resources==5.12.0',
        'k-diffusion',
        'laion_clap',
        'nwt-pytorch',
        'pandas',
        'pedalboard',
        'prefigure',
        'pytorch_lightning==1.7.4', 
        'PyWavelets',
        'sentencepiece',
        's3fs',
        'torch',
        'torchaudio',
        'torchmetrics==0.11.4',
        'tqdm',
        'transformers==4.30.2',
        'v-diffusion-pytorch',
        'vector-quantize-pytorch',
        'wandb',
        'webdataset',
        'x-transformers'
    ],
)